#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

from typing import List, Sequence
import math
import numbers
import bisect
import itertools


def gcd_many(s: List[int]) -> int:
    '''
    Summary: 求一组数字的 最大公约数
    '''
    g = 0
    for i in range(len(s)):
        if i == 0:
            g = s[i]
        else:
            # math.gcd 可以求2个数字的最大公约数
            g = math.gcd(g, s[i])
    return g


def flat_and_sort_list(ls: List) -> list:
    '''列表降维 + 排序'''
    ls = list(itertools.chain(*ls))
    ls.sort()
    return ls


def find_x(x: numbers.Number,
           y_list: List[numbers.Number],
           use_smaller: bool = False,
           use_larger: bool = False,
           func=None,
           maximum: numbers.Number = None,
           minium: numbers.Number = None):
    '''
    从list中找到 参数1 最接近的值，或者小于/大于的值
    Params:
        - x: numbers.Number, 原始值
        - y_list: List[numbers.Number], 列表
        - use_smaller: bool = False, True则找到最接近的比原始值小的值
        - use_larger: bool = False, True 则找到最接近的比原始值大的值
        - func: bool = False, True 则采用 fun(x)的值作为对比
        - maximum: numbers.Number = None, 提供的话则限制最大值不超过此值
        - minium: numbers.Number = None, 提供则限制最小值不小于此值
    '''
    if not isinstance(x, numbers.Number):
        raise TypeError('find_x 只能处理数字')
    if not isinstance(y_list, Sequence) or not y_list:
        raise TypeError('y_list 应为List[float]')
    y_list.sort()  # 进行正向排序
    # 如果限定了最大值(限定在 y_list之间才有意义)
    if maximum and maximum >= y_list[0] and maximum <= y_list[-1]:
        y_list = y_list[:bisect.bisect(y_list, maximum)]  # 切片到<=maximum的值
    # 如果限定了最小值(限定在 y_list之间才有意义)
    if minium and minium >= y_list[0] and minium <= y_list[-1]:
        y_list = y_list[bisect.bisect_left(y_list, minium):]  # 切片到>=maximum的值
    # 寻找合适额值
    if x in y_list:
        return x
    # 如果比最小值还小
    elif x < y_list[0]:
        return y_list[0]
    elif x > y_list[-1]:
        return y_list[-1]
    else:
        index = bisect.bisect(y_list, x)
        small = y_list[index-1]
        large = y_list[index]
        if use_smaller:  # 如果需要找到小的
            return small
        elif use_larger:  # 如果需要找到大的
            return large
        else:  # 根据条件找到最靠近的
            if not func:  # 是否使用平方数代替直接的对比（考虑到管径选择时候选用直径平方更加合适 math.pow）
                return small if (small+large)/2 >= x else large  # 对应的离子强度标定
            else:
                return small if (func(small) + func(large))/2 >= func(x) else large



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2020-05-09

'''
温度
'''
from dataclasses import dataclass
# app
from .fomulas import get_water_dynamic_viscosity, get_temp_correction_factor

'''
[温度相关系数 经验公式]
1) pk1 : [H][HCO3] == [H2CO3]
2) pk2 : [H][CO3] == [HCO3]
3) pkw : [H][OH] == [H2O]
4) pkh : ??
5) pks : ?? (CaCO3)
'''


@dataclass
class Temp:
    temp: float = 25  # 默认为摄氏度 非绝对温度

    @property
    def K(self) -> float:
        return self.temp + 273.15

    @property
    def F(self) -> float:
        '''
        Summary: 华氏温度
        '''
        return self.temp*1.8+32

    @property
    def RA(self) -> float:
        '''
        Summary: 兰氏度Rankine
        '''
        return self.temp * 1.8 + 32 + 459.67

    @property
    def R(self) -> float:
        '''
        Summary: 列氏度Réaumur
        '''
        return self.temp * 0.8

    @property
    def pk1(self) -> float:
        # return 3447/self.K-15.08 + 0.0331*self.K
        return 6.38

    @property
    def pk2(self) -> float:
        # return 2929/self.K - 6.65 + 0.0240 * self.K
        return 10.21

    @property
    def pkw(self) -> float:
        return 4470/self.K - 6.09 + 0.0171*self.K

    @property
    def pkh(self) -> float:
        return -2218/self.K + 12.7 - 0.0127*self.K

    @property
    def pks(self) -> float:
        return 8.03 + 0.01183*self.K

    @property
    def k1(self) -> float:
        return 10**(-self.pk1)

    @property
    def k2(self) -> float:
        return 10**(-self.pk2)

    @property
    def kw(self) -> float:  # 稀溶液中水的kw
        if self.temp == 25:
            return 1e-14
        return 10**(-self.pkw)

    @property
    def kh(self) -> float:
        return 10**(-self.pkh)

    @property
    def ks(self) -> float:
        return 10**(-self.pks)

    @property
    def pka_nh4(self) -> float:
        '''NH4+ <--> NH3 + H+ , pKa = 9.24'''
        return 9.24

    @property
    def ka_nh4(self) -> float:
        '''NH4+ <--> NH3 + H+ , pKa = 9.24'''
        return 10**(-self.pka_nh4)

    @property
    def kb_nh3(self) -> float:
        '''NH3 + H2O <--> NH4+ + OH- , pKa = 4.76
        推导得到： Ka = Kw/Kb
        '''
        return self.kw / self.ka_nh4

    @property
    def water_dynamic_viscosity(self) -> float:
        '''
        Summary: 水的粘度与温度关系 ,Pa*s
        '''
        return get_water_dynamic_viscosity(self.temp)

    @property
    def temp_correction_factor(self) -> float:
        '''
        Summary: 温度补偿系数
        '''
        return get_temp_correction_factor(self.temp)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

'''
Summary: Time
'''
from datetime import datetime
from functools import lru_cache


@lru_cache(maxsize=128, typed=False)
def utc2local(utc_dtm):
    # UTC 时间转本地时间（ +8:00 ）
    local_tm = datetime.fromtimestamp(0)
    utc_tm = datetime.utcfromtimestamp(0)
    offset = local_tm - utc_tm
    return utc_dtm + offset


@lru_cache(maxsize=128, typed=False)
def local2utc(local_dtm):
    # 本地时间转 UTC 时间（ -8:00 ）
    return datetime.utcfromtimestamp(local_dtm.timestamp())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

'''
Summary: Dict
'''
import copy


def deep_update_dict(d: dict, u: dict) -> dict:
    # 用于深度 update  dict
    d = copy.deepcopy(d)
    if not u:
        return d
    for k, v in u.items():
        if isinstance(v, dict):
            r = deep_update_dict(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

'''
Summary: Str
'''
import string
import random


def generate_random_str(size: int = 6, chars=None, only_number=False):
    '''
    Summary: 生成随机字符串
    '''
    if not chars:
        if not only_number:
            chars = string.ascii_letters + string.digits
        else:
            chars = string.digits
    return ''.join(random.choice(chars) for _ in range(size))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2020-05-09

'''
Summary: PH相关
'''

from dataclasses import dataclass
from typing import Optional
import math
import copy
# app
from .temp import Temp
from .water_utils import h_mol_fx_A, h_mol_fx_ph


@dataclass
class PhDc:
    '''
    PH class, 包含一系列ph相关计算的类
    OH 和 H 看似 可以通过H*OH = KW 限制
    其实不然
    需要以 H-OH 的差值作为唯一依据
        - 质子平衡时 不考虑 活度
        - K1 K2 平衡需要考虑活度
        - H*OH = kw (其实等于  H_active * OH_active = kw)
        - H和OH都是abs(valence) = 1 
        - 因此 kw / activity_k**2 = H*OH
        - 单位都是 mol/l
    '''
    temp: float = 25  # 温度
    y1: float = 1  # 1价活度
    T: Temp = None
    h: float = 0  # 最终 h 实际质子平衡
    oh: float = 0  # 最终oh_mol 实际质子平衡
    # 计算
    y1: float = None  # 1价离子 活度
    a0a1a2: dict = None  # 碳酸平衡比例
    A: Optional[float] = None  # A
    ph: Optional[float] = None  # ph

    def __post_init__(self):
        if not self.T:
            self.T = Temp(self.temp)
        self.set_by_ph()
        self.set_by_A()
        self.set_a0a1a2()

    def set_by_ph(self):
        # 通过ph设定
        if self.ph is None:
            return None
        self.h = h_mol_fx_ph(ph=self.ph, y1=self.y1)
        self.oh_active = self.T.kw / self.h_active
        self.A = self.h-self.oh

    def set_by_A(self):
        if self.A is None:
            return None
        self.h = h_mol_fx_A(A=self.A, kw=self.T.kw, y1=self.y1)
        self.oh_active = self.T.kw / self.h_active
        self.ph = -math.log10(self.h_active)

    @property
    def h_active(self) -> float:
        '''考虑活度的[h]'''
        return self.h * self.y1

    @h_active.setter
    def h_active(self, val: float):
        '''考虑活度的[h]'''
        self.h = val / self.y1

    @property
    def oh_active(self) -> float:
        '''考虑活度的[oh]'''
        return self.oh * self.y1

    @oh_active.setter
    def oh_active(self, val: float):
        '''考虑活度的[h]'''
        # val = validate_numbers(val)
        self.oh = val / self.y1

    def set_a0a1a2(self):
        '''
        Summary: 设置co2 hco3 co3 比例,初始化时候运行
        Depend : self.ph
        Note   : 如果传入实际的ct，应该考虑 co2溢出的可能，因此比例增加一个co2gas(ag)
        Return : dict
        '''
        h = self.h_active
        k1 = self.T.k1
        k2 = self.T.k2
        ka_nh4 = self.T.ka_nh4
        if self.h >= 1:
            nh3 = 0
        elif self.oh >= 1:
            nh3 = 1
        else:
            nh3 = ka_nh4/(h + ka_nh4)
        nh4 = 1 - nh3
        # 不考虑活度
        # a0 = (h**2) / (h**2 + k1 * h + k1 * k2)
        a1 = (k1*h) / (h**2 + k1 * h + k1 * k2)
        a2 = k1*k2 / (h**2 + k1 * h + k1 * k2)
        a0 = 1-a1-a2
        ag = 0
        self.a0a1a2 = dict(a0=a0, a1=a1, a2=a2, ag=ag, nh3=nh3, nh4=nh4)

    def a0a1a2_with_ct(self, ct: float) -> dict:
        '''
        Summary: 根据当前的a0a1a2比例 传入ct，不考虑co2气体溢出可能
        Depend : self.a0a1a2
        Return : dict，实际的 co2g co2 hco3 co3
        para1  : ct： 总C mmol_l
            - return的单位与 传入ct一致
        '''
        aaa = copy.copy(self.a0a1a2)  # 复制一份
        aaa['ag'] = 0  # co2气体部分
        a0, a1, a2 = aaa['a0'], aaa['a1'], aaa['a2']
        dt = {}
        # if a0*ct > 10:
        #     new_ct = 10/a0  # new ct
        # else:
        #     new_ct = ct
        new_ct = ct
        dt['co2'] = new_ct*a0
        dt['hco3'] = new_ct*a1
        dt['co3'] = new_ct*a2
        # 增加 'ag'部分，在 H 平衡时，也要考虑此部分包含的质子
        dt['ag'] = ct-new_ct  # 此部分转换为 ag了
        dt['h_in_ct'] = (dt['co2']+dt['ag'])*2 + dt['hco3']
        return dt

    @property
    def h_in_h2co3_k(self):
        '''碳酸体系 包含 H'''
        return self.a0a1a2['a0'] * 2 + self.a0a1a2['a1']

    @property
    def h_in_nh3_k(self) -> float:
        '''氨体系 包含的 H: 仅 nh4 包含 H '''
        return self.a0a1a2['nh4']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08


'''
Summary: Class decorator
'''
import functools
import time


def clock(func):
    '''
    计时器，用于计算函数的运行时间
    bug: 接收 lambda 作为参数发生错误
    '''
    @functools.wraps(func)
    def wraps(*args, **kwargs):
        t1 = time.perf_counter()
        result = func(*args, **kwargs)  # 被装饰的func 传入所有参数
        time_delt = time.perf_counter() - t1
        # print(f'{func.__name__}耗时{time_delt}秒')
        print(
            {'func': func.__name__, 'time': time_delt}
        )
        return result
    return wraps


class classproperty(object):
    '''用于将classmethod 装饰为 property'''

    def __init__(self, func):
        self.func = classmethod(func)

    def __get__(self, obj=None, type=None):
        return self.func.__get__(obj, type)()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# File    :   names.py
# Time    :   2019/01/25 00:06:46
# Author  :   Jack Li
# Contact :   allanth3@163.com

'''
Ions Names Dict
完全继承Dict

手动加上 加减乘除 4个方法
'''

import copy
import numbers
from collections import Counter


class MyCounter(Counter):
    '''
    Summary: 以namedtuple 作为key 的dict
    Note   : 实现dict的基本数学运算
    '''

    @property
    def all_names(self):
        '''
        Summary: 所有namedtuple的name属性list
        Note   : 有普通的key则返回空列表
        Return : list
        '''
        try:
            return [x.name for x in self.keys()]
        except:
            return []

    @classmethod
    def from_dict(cls, dt):
        '''
        从普通dt生成MyCounter()
        key小写化，去除v 非数字或0
        **True 是 numbers.Real
        '''
        new = {str(k).lower(): v for k,
               v in dt.items() if v and isinstance(v, numbers.Real) and not isinstance(v, bool)}
        return cls(new)

    def __imul__(self, i: float):
        '''self *= i'''
        self = MyCounter({k: v*i for k, v in self.items()})

    def __mul__(self, i: float):
        '''self * i'''
        new = {k: v*i for k, v in self.items()}
        return MyCounter(new)

    def __rmul__(self, i: float):
        '''i * self'''
        return self.__mul__(i)

    def __truediv__(self, i: float):
        '''self / i'''
        new = {k: v/i for k, v in self.items()}
        return MyCounter(new)

    def __itruediv__(self, i: float):
        '''self /= i '''
        self = MyCounter({k: v/i for k, v in self.items()})

    def __add__(self, other):
        '''like Counter __add__ but return MyCounter'''
        #cls = type(self)
        if other is None:
            return copy.copy(self)
        return MyCounter(super().__add__(other))

    def __iadd__(self, other):
        '''like Counter __iadd__ but return MyCounter'''
        #cls = type(self)
        if other is None:
            return copy.copy(self)
        return super().__iadd__(other)

    def __sub__(self, other):
        '''like Counter __sub__ but return MyCounter'''
        #cls = type(self)
        if other is None:
            return copy.copy(self)
        return MyCounter(super().__sub__(other))

    def __isub__(self, other):
        '''like Counter __isub__ but return MyCounter'''
        #cls = type(self)
        if other is None:
            return copy.copy(self)
        return super().__isub__(other)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

'''
Summary: Json file
'''
import os
import json


def get_from_json(filename: str, dirpath: str) -> dict:
    '''
    Summary: 从json文件中获得默认chem data
    NOTE : open 的 'r' 只读模式必须判断是否存在文件
    '''
    file = f'{filename}.json'
    json_path = os.path.join(dirpath, file)
    if not os.path.exists(json_path):
        print(f'{file} 不存在')
        return {}
    with open(json_path, 'r', encoding='utf-8') as f:
        st = f.read()
        if st:
            return json.loads(st)
        else:
            return {}


def save_to_json(data: dict, filename: str, dirpath: str):
    '''
    Summary: 将修改过的json内容保存到自定义的json文件
    NOTE : open 的 'w' 如果文件不存在会自己创建
    '''
    try:
        file = f'{filename}.json'
        js = json.dumps(data, ensure_ascii=False)  # 生成json str
        json_path = os.path.join(dirpath, file)
        with open(json_path, 'w', encoding='utf-8') as f:
            f.write(js)
    except Exception as e:
        print(e)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2020-05-08

'''
与Water相关的  Function
'''
from sympy import symbols, solve
import math
from typing import Any

'''
===================以下为线性公式，适合大量数据计算======================
格式 f(x)  : f_fx_x()
浓度单位默认都为mol/L
'''


def get_ion_strength(mmol_l: float, valence: float):
    '''
    Summary: 求单个离子ion strength
    para1  : mmol_l
    para2  : valence 价位
    '''
    return 0.5 * mmol_l/1000 * (valence**2)


def get_h_active(h: float = None,
                 h_active: float = None,
                 y1: float = None) -> float:
    if h_active:
        return h_active
    return h*y1


def ph_fx_h(h: float, y1: float = 1) -> float:
    '''
    Summary: h求ph
    '''
    return -math.log10(h*y1)


def h_mol_fx_ph(ph: float, y1: float = 1) -> float:
    '''
    Summary: 根据ph计算 h mol/L
    Note: 
        通过公式计算得到的为 h_active, 实际h需要考虑y1
    '''
    return 10**(-ph) / y1


def oh_mol_fx_ph(ph: float,
                 kw: float = 1e-14,
                 y1: float = 1) -> float:
    '''
    Summary: oh(ph)
        y1 为一价盐活度
    '''
    return kw/y1**2/(10**-ph)


def h_mol_fx_A(A: float,
               kw: float = 1e-14,
               y1: float = 1) -> float:
    '''
    Summary: h(A) 根据 质子差值 计算 H mol/l
    Note   : ct=0,nt=0
        - y1 1加盐活度系数
    '''
    # kw = kw/y1**2
    A = A * y1
    if A >= 0:  # 酸性
        h = (-A+math.sqrt(A**2 - 4*-kw))/2 + A
    else:  # 碱性
        A = abs(A)
        oh = (-A+math.sqrt(A**2 - 4*-kw))/2 + A
        h = kw/oh
    return h/y1


def ph_fx_A(A: float,
            kw: float = 1e-14,
            y1: float = 1) -> float:
    '''
    Summary: ph(h_sub_oh_mol) 根据A 计算ph
    A= h_sub_oh_mol
        - y1 为1加盐活度系数
    '''
    h = h_mol_fx_A(A=A, kw=kw, y1=y1)
    return -math.log10(h)


def a0_fx_h_mol(h: float = None,
                h_active: float = None,
                k1: float = 4.168693834703355e-07,
                k2: float = 6.16595001861481e-11,
                y1: float = 1) -> float:
    '''
    Summary: a0(h)  # H2CO3 比例
        - 返回之前把h 转换为 h_active
    '''
    h = get_h_active(h=h, h_active=h_active, y1=y1)
    return (h**2) / (h**2 + k1 * h + k1 * k2)


def a1_fx_h_mol(h: float = None,
                h_active: float = None,
                k1: float = 4.168693834703355e-07,
                k2: float = 6.16595001861481e-11,
                y1: float = 1) -> float:
    '''
    Summary: a1(h)  # HCO3- 比例
    '''
    h = get_h_active(h=h, h_active=h_active, y1=y1)
    return (k1*h) / (h**2 + k1 * h + k1 * k2)


def a2_fx_h_mol(h: float = None,
                h_active: float = None,
                k1: float = 4.168693834703355e-07,
                k2: float = 6.16595001861481e-11,
                y1: float = 1) -> float:
    '''
    Summary: a2(h)  # CO3-2 比例
    '''
    h = get_h_active(h=h, h_active=h_active, y1=y1)
    return k1*k2 / (h**2 + k1 * h + k1 * k2)


def a0_fx_ph(ph: float,
             k1: float = 4.168693834703355e-07,
             k2: float = 6.16595001861481e-11) -> float:
    '''
    Summary: a0(ph)  # H2CO3 比例，无量纲
    '''
    h = 10**(-ph)  # 即为 h_active
    return (h**2) / (h**2 + k1 * h + k1 * k2)


def a1_fx_ph(ph: float,
             k1: float = 4.168693834703355e-07,
             k2: float = 6.16595001861481e-11) -> float:
    '''
    Summary: a1(ph)  # HCO3- 比例，无量纲
    '''
    h = 10**(-ph)  # 即为 h_active
    return (k1*h) / (h**2 + k1 * h + k1 * k2)


def a2_fx_ph(ph: float,
             k1: float = 4.168693834703355e-07,
             k2: float = 6.16595001861481e-11) -> float:
    '''
    Summary: a2(ph)  # CO3-2 比例，无量纲
    '''
    h = 10**(-ph)  # 即为 h_active
    return k1*k2 / (h**2 + k1 * h + k1 * k2)


def a_nh3_fx_ph(ph: float,
                ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: nh3(ph)  # NH3 比例，无量纲
    '''
    h = 10**(-ph)  # 即为 h_active
    return ka_nh4/(h+ka_nh4)


def a_nh4_fx_ph(ph: float,
                ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: nh4(ph)  # NH4+ 比例，无量纲
    '''
    h = 10**(-ph)  # 即为 h_active
    return h/(h+ka_nh4)


def h_in_ct_k_fx_h_mol(h: float = None,
                       h_active: float = None,
                       kw: float = 1e-14,
                       k1: float = 4.168693834703355e-07,
                       k2: float = 6.16595001861481e-11,
                       y1: float = 1) -> float:
    '''
    Summary: h(A)，碳酸平衡中h  mol/l 占 ct 比例
    Note   : ct=0,nt=0
    '''
    a0 = a0_fx_h_mol(h=h, h_active=h_active, k1=k1, k2=k2, y1=y1)
    a1 = a1_fx_h_mol(h=h, h_active=h_active, k1=k1, k2=k2, y1=y1)
    return a0*2 + a1


def a_nh3_fx_h_mol(h: float = None,
                   h_active: float = None,
                   ka_nh4: float = 5.754399373371567e-10,
                   y1: float = 1) -> float:
    '''
    Summary: nh3(h)  # NH3-2 比例
    '''
    h = get_h_active(h=h, h_active=h_active, y1=y1)
    return ka_nh4/(h+ka_nh4)


def a_nh4_fx_h_mol(h: float = None,
                   h_active: float = None,
                   ka_nh4: float = 5.754399373371567e-10,
                   y1: float = 1) -> float:
    '''
    Summary: nh4(h)  # NH4+ 比例
    '''
    h = get_h_active(h=h, h_active=h_active, y1=y1)
    return h/(h+ka_nh4)


def cal_h_balance_deviation(h: float,
                            A: float,  # 非活性
                            ct: float = 0,
                            nt: float = 0,
                            kw: float = 1e-14,
                            k1: float = 4.168693834703355e-07,
                            k2: float = 6.16595001861481e-11,
                            ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: 
        - 水中 h,A,ct,nt,kw,k1,k2,ka_nh4 具备 质子平衡，最终可以列出如下5次方程
        - 解方程消耗太大，因此可以反向进行求和
        - 用于 近似法求h_mol使得方程最接近0 (abs(result)<= 1e-7)
        - return 误差值(通过* 1e40放大）
    '''
    k3 = ka_nh4
    h = h
    # CT and NT ,5次方程
    x1 = k3+nt-A
    x2 = -kw-A*k3
    a = 1
    b = 2*ct+k1+x1
    c = 2*ct*k3 + k1*ct + k1*k2 + k1*x1 + x2
    d = k1*k3*ct + k1*k2*x1 + x2*k1 - kw*k3
    e = x2*k1*k2 - k1*k3*kw
    f = -k1*k2*k3*kw
    deviation = a*h**5 + b*h**4 + c*h**3 + d*h**2 + e*h + f
    return deviation*1e40


def h_mol_fx_A_CT_NT(A: float,
                     ct: float = 0,
                     nt: float = 0,
                     kw: float = 1e-14,
                     k1: float = 4.168693834703355e-07,
                     k2: float = 6.16595001861481e-11,
                     ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: A, CT, NT 单位mol/L
    Note   : 最终要解1元3次方程，肯定有一个实数解
    Return :
    para1  : A : 总的 质子： h_in_ct + h_in_nt + [H]-[OH] mol/L
    para2  : ct: 总C  mol/L
    para3  : ct: 总NH3-NH4  mol/L
    '''
    k3 = ka_nh4
    h = symbols('h')
    # CT and NT ,5次方程
    x1 = k3+nt-A
    x2 = -kw-A*k3
    a = 1
    b = 2*ct+k1+x1
    c = 2*ct*k3 + k1*ct + k1*k2 + k1*x1 + x2
    d = k1*k3*ct + k1*k2*x1 + x2*k1 - kw*k3
    e = x2*k1*k2 - k1*k3*kw
    f = -k1*k2*k3*kw
    # 推导过程见照片
    ans = solve(
        a*h**5 + b*h**4 + c*h**3 + d*h**2 + e*h + f,
        h
    )
    # 取实数、正数解
    for x in ans:
        try:
            x = float(x)
            if x > 0:
                return x
        except:
            continue
    return None


def h_mol_fx_A_CT_NT_binary(A: float,
                            ct: float = 0,
                            nt: float = 0,
                            kw: float = 1e-14,
                            k1: float = 4.168693834703355e-07,
                            k2: float = 6.16595001861481e-11,
                            ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: 二分法尝试计算
    Note   : 最高ph范围 = -2 - 16, h 范围 = 1e-16 ~ 1e-2 
    Return : h
    para1  : A : 总的 质子： h_in_ct + h_in_nt + [H]-[OH] mol/L
    para2  : ct: 总C  mol/L
    para3  : ct: 总NH3-NH4  mol/L
    '''
    h_range = range(-2, 17)
    h_res = None
    i_res = None
    # 先找到大的ph范围（1--10）**10**i，使 方程式正好为﹣
    # 在此基础上再找前缀
    # 此时前缀范围1-10能逐渐接近实际解
    for i in h_range:  # 极限19次
        h = 10**(-i)
        deviation = cal_h_balance_deviation(
            h=h, A=A, ct=ct, nt=nt, kw=kw, k1=k1, k2=k2, ka_nh4=ka_nh4)
        # print(f'1:i={i},h={h},deviation={deviation}')
        if deviation == 0:
            i_res = i
            h_res = h
            return h_res
        if deviation < 0:
            i_res = i
            h_res = h
            break

    start, end = 1, 10
    i = 0
    while i <= 15:  # 极限16次
        first = (start+end)/2
        h1 = first * h_res  # first * 1e-i
        deviation = cal_h_balance_deviation(
            h=h1, A=A, ct=ct, nt=nt, kw=kw, k1=k1, k2=k2, ka_nh4=ka_nh4)
        if deviation == 0:
            return h1
        elif deviation < 0:
            # 说明 first 太小
            start, end = first, end
        else:
            start, end = start, first
        i += 1
    return h1


def h_mol_fx_A_CT(A: float,
                  ct: float = 0,
                  kw: float = 1e-14,
                  k1: float = 4.168693834703355e-07,
                  k2: float = 6.16595001861481e-11) -> float:
    '''
    Summary: A, CT 单位mol/L
    Note   : 最终要解1元4次方程
    Return :
    para1  : A : 总的 质子： h_in_h2co3 + [H]-[OH] mol/L
    para2  : ct: 总C  mol/L

    (-h_sub_oh+math.sqrt(h_sub_oh**2 - 4*-1e-14))/2 + h_sub_oh
    '''

    h = symbols('h')
    # CT, 4次方程
    a = 1
    b = 2*ct - A + k1
    c = k1*ct - kw - k1*A+k1*k2
    d = -k1*kw - k1*k2*A
    e = -k1*k2*kw
    # 推导过程见照片
    ans = solve(
        a*h**4 + b*h**3 + c*h**2 + d*h + e,
        h
    )
    # 取实数、正数解
    for x in ans:
        try:
            x = complex(x)
            if x.real > 0:
                return abs(x)
        except:
            continue
    return None


def h_mol_fx_A_NT(A: float,
                  nt: float = 0,
                  kw: float = 1e-14,
                  ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: A, NT 单位mol/L，ct = 0 时候 求平衡 h
    Note   : 最终要解1元3次方程
    Return :
    para1  : A : 总的 质子： h_in_nt + [H]-[OH] mol/L
    para2  : nt: 总Nh3-NH4  mol/L

    (-h_sub_oh+math.sqrt(h_sub_oh**2 - 4*-1e-14))/2 + h_sub_oh
    '''
    k3 = ka_nh4
    h = symbols('h')
    # CT, 4次方程
    a = 1
    b = nt+k3-A
    c = -kw - A*k3
    d = -k3*kw
    # 推导过程见照片
    ans = solve(
        a*h**3 + b*h**2 + c*h + d,
        h
    )
    # 取实数、正数解
    for x in ans:
        try:
            x = complex(x)
            if x.real > 0:
                return abs(x)
        except:
            continue
    return None


def get_h_mol_by_A_CT_NT(A: float,
                         ct: float = 0,
                         nt: float = 0,
                         kw: float = 1e-14,
                         k1: float = 4.168693834703355e-07,
                         k2: float = 6.16595001861481e-11,
                         ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: 根据不同情况 CT,NT 有零项时，选用4次或3次方程求解(严格按照解方程)
    Note   : 
    Return : h mol/L
    para1  : A : 总的 质子： h_in_ct + h_in_nt + [H]-[OH] (mol/L)
    para2  : ct: 总C  mol/L
    para3  : ct: 总NH3-NH4  mol/L
    '''
    if not ct and not nt:
        # 仅h，oh x**2
        return h_mol_fx_A(A=A, kw=kw)
    if ct and not nt:
        # h,oh,a0,a1,a2,x**4
        return h_mol_fx_A_CT(A=A, ct=ct, kw=kw, k1=k1, k2=k2)
    if not ct and nt:
        # h,oh,a0,a1,a2,x**3
        return h_mol_fx_A_NT(A=A, nt=nt, kw=kw, ka_nh4=ka_nh4)
    if ct and nt:
        # h,oh,a0,a1,a2,a_nh3,a_nh4,x**5
        return h_mol_fx_A_CT_NT(A=A, ct=ct, nt=nt, kw=kw, k1=k1, k2=k2, ka_nh4=ka_nh4)


def get_h_mol_by_A_CT_NT_binary(A: float,
                                ct: float = 0,
                                nt: float = 0,
                                kw: float = 1e-14,
                                k1: float = 4.168693834703355e-07,
                                k2: float = 6.16595001861481e-11,
                                ka_nh4: float = 5.754399373371567e-10) -> float:
    '''
    Summary: 根据不同情况 CT,NT 有零项时，选用4次或3次方程求解(二分法)
    Note   : 
    Return : h mol/L
    para1  : A : 总的 质子： h_in_ct + h_in_nt + [H]-[OH] (mol/L)
    para2  : ct: 总C  mol/L
    para3  : ct: 总NH3-NH4  mol/L
    '''
    if not ct and not nt:
        # 仅h，oh x**2
        return h_mol_fx_A(A=A, kw=kw)
    else:
        return h_mol_fx_A_CT_NT_binary(A=A, ct=ct, nt=nt, kw=kw, k1=k1, k2=k2, ka_nh4=ka_nh4)


def get_osp(mmol_l: float, temp: float = 25) -> float:
    '''
    Summary: 渗透压计算
    '''
    return mmol_l * 8.314 * (temp + 273.15) / 100000



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2020-05-02

'''
Summary: 物理公式及计算
** 仅复杂计算采用lru_cache
'''
from typing import List
import math
from functools import lru_cache


def get_round_area(diameter: float) -> float:
    '''直径计算圆面积'''
    return (diameter/2)**2*math.pi


def get_round_radius(area: float) -> float:
    '''获得半径'''
    return (area/math.pi)**(1/2)


def get_round_diameter(area: float) -> float:
    '''面积计算圆直径'''
    return (area/math.pi)**(1/2)*2


def get_round_tank_size(volumn: float, height_diameter: float = 1.2) -> dict:
    '''
    根据有效体积V，超高H_add,高径比(有效高度:直径) 计算size
    '''
    # diameter = height / height_diameter
    # R = diameter/2 = height/height_diameter/2
    # area = R**2*math.pi
    # area = (height/height_diameter/2)**2 * math.pi
    # volumn = area * height = (height/height_diameter/2)**2 * math.pi * height
    # height**3 = volumn / math.pi * (height_diameter*2)**2
    height = (volumn / math.pi * (height_diameter*2)**2)**(1/3)
    diameter = height/height_diameter
    return {'height': height, 'diameter': diameter}


def get_round_tank_height(volumn: float, diameter: float) -> float:
    '''根据V,D得到H'''
    area = get_round_area(diameter=diameter)
    return volumn/area


def get_round_tank_volumn(height: float, diameter: float) -> float:
    '''根据D，H获得 圆形槽体V'''
    return height*(diameter/2)**2*math.pi


def get_v_round_pip(m3ph: float, diameter: float) -> float:
    '''
    Summary: 圆管计算流速
    Params:
        - q: 流量 m3/h 
        - diameter: 直径 m
        return 单位m/s
    '''
    s = get_round_area(diameter=diameter)
    return m3ph / s / 3600


def get_temp_correction_factor(temp: float) -> float:
    '''
    Summary: 温度补偿系数计算
        采用国际水和水蒸气性质协会(IAPWS)官方公式
        DOW也采用此公式作为温度补偿计算
    '''
    k = temp+273.15  # 热力学温度
    kk = k/300
    a1 = 280.68*(kk**(-1.9))
    a2 = 511.45*(kk**(-7.7))
    a3 = 61.131*(kk**(-19.6))
    a4 = 0.45903*(kk**(-40))
    return 890/(a1+a2+a3+a4)


def get_hydro_d(area: float, perimeter: float) -> float:
    '''
    Summary: 根据 截面积 和 湿润周长，计算 水力直径（计算re）
    '''
    return area*4 / perimeter


def get_water_viscosity(temp: float) -> float:
    '''
    Summary: 计算 水的粘度 N·s/㎡ = Pa*s
    '''
    t = temp  # + 273.15
    return 0.001779/(1+0.03368*t+0.0002210*t**2)


def get_water_dynamic_viscosity(temp: float) -> float:
    '''
    Summary: 根据 温度 计算 水的 运动粘度
    单位: Pa*s 
    '''
    return 0.001779 / (1+0.03368*temp+0.0002210*temp**2)


def get_water_kinematic_viscosity(temp: float) -> float:
    '''
    Summary: 根据 温度 计算 水的 运动粘度
    单位: Pa*s 
    '''
    dynamic_viscosity = 0.001779 / (1+0.03368*temp+0.0002210*temp**2)
    density = 997
    return dynamic_viscosity / density


def get_re(d: float,  # 4倍面积/润湿的周长
           v: float,
           density_kg_m3: float = 997,  # 密度 kg/m3
           dynamic_viscosity: float = None,  # 动力粘度 Pa*s
           kinematic_viscosity: float = 1.006e-6  # 运动粘度 m2/s
           ) -> float:
    '''
    Summary: 雷诺数计算 Reynolds number
        - Re=ρvd/η (1)
        - d: 定型尺寸 m (圆管为直径) 
        - ρ：density 密度25℃水为997, kg/m3
        - v: 流速 m/s 
        - η: dynamic_viscosity 动力粘度 Pa*s 或 kg/(m*s) 或 N·s/㎡:
            在流体中取两面积各为1m2，相距1m，
            相对移动速度为1m/s时所产生的阻力称为动力粘度。
            单位Pa.s（帕.秒） [2]  。
            过去使用的动力粘度单位为泊或厘泊，泊（poise）或厘泊为非法定计量单位。
            20℃时水的运动黏度ν=1.006×10^(-6) (m2/s)
        - υ:kinematic_viscosity (m2/s)
            - η＝ρυ -> η = kg/m3 * m2/s = kg/(m*s) = Pa·s 即动力粘度的单位
                - 压强的定义Pa=N/m^2,
                - Pa*s=N*s/m^2
                - 牛顿第二定律N=kg*m/s^2,
                - 所以 Pa*s=(kg*m/s^2)*s/m^2=kg/(m*s)
        - 式中的动力粘度η用运动粘度υ来代替，因η＝ρυ, 因此 Re = vd/υ (2)#
            - Re = 流速*定型尺寸/运动粘度 
            - η＝ρυ -> η = kg/m3 * m2/s = kg/(m*s) = Pa·s 即动力粘度的单位
            - η/ρ＝υ
    Params:
        - d: 定型尺寸 m (圆管为直径)
        - v: 流速 m/s 
        - density_kg_m3: 密度 kg/m3
        - dynamic_viscosity: 动力粘度, Pa·s， 25℃水=0.89e-3 Pa·s (kg/(m·s))
        - kinematic_viscosity: 运动粘度, m2/s,25℃时水的运动粘度0.893e-6 (m2/s)
    Unit:
        - m * m/s / (m2/s) = 无量纲
        - m * m/s * kg/m3 / (kg/(m*s)) => kg/m/s / (kg/m/s) -> 1 
    '''
    if kinematic_viscosity:
        # 运动粘度计算
        return d*v/kinematic_viscosity
    # 粘度计算
    return d*v*density_kg_m3 / dynamic_viscosity


def get_re_water(d: float, v: float, temp: float):
    '''
    Summary: 水的 雷诺数
    '''
    dynamic_viscosity = get_water_dynamic_viscosity(temp=temp)
    return get_re(d=d, v=v, dynamic_viscosity=dynamic_viscosity)


def get_dp_hazen_williams_law(re: float, v_m_s: float, d_m: float, length_m: float) -> float:
    '''
    Summary: Hazen williams law
            ---H=10.67 * 121**-1.85 * D**-4.87 * Q*1.85 L 121指PVC管道的表面粗糙度常数

            hf = ru *(l/D)*(v^2/2/g)
            - ru: 沿程阻力系数
            - l :管道长度 m
            - D : 管道内径 m, 用水力半径替代
            - v : 平均流速 m/s
            - g : 重力加速度 m/s2
    '''
    # 返回为 bar 单位
    g = 9.801
    if re < 2320:
        ru = 64/re
    elif 3000 < re < 10e5:
        ru = 0.3164*re**(-0.25)
    else:
        ru = 0.308*(0.842-math.log(re))**-2
    return ru*length_m/d_m*v_m_s**2/(2*g) / 10


@lru_cache(maxsize=128)
def debye_huckel(ion_strength: float = 0, valence: float = 1) -> float:
    '''
    Summary: 德拜-休克尔公式：适用于【强电解质】【稀溶液】的活度计算经验公式
            25℃，A=0.509(mol/kg)**(-0.5)
    Note   : lnx= -A * zi**2 * I**0.5
    Return : ya
    para1: ion_strength: 离子强度
    para2: valence:离子价位
    '''
    i = ion_strength
    # res = (math.e)**(-0.509*valence**2 *
    #                  i**0.5)
    res = (math.e)**(-0.509*valence**2 * (i**0.5/(1+i**0.5)-0.3*i))
    return round(res, 3)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author : Jack Li
# Email  : allanth3@163.com
# Date   : 2021-06-08

'''
Summary:number
'''


def validate_numbers(val: float, ge_zero: bool = True, throw: bool = False, default: float = 0):
    '''
    验证传入值是否数字，并且符合要求
    ge_zero: 必须 >= 0
    throw:抛出错误，如果设置为False  则返回 default
    '''
    try:
        val = float(val)
    except:
        if throw:
            raise TypeError('val 必须是数字')
        else:
            return default
    if ge_zero:
        if val < 0:
            if throw:
                raise ValueError(f'Value{val} < 0')
            else:
                return default
    return val



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['jackutils']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.8.2,<2.0.0', 'sympy>=1.8,<2.0']

setup_kwargs = {
    'name': 'jackutils',
    'version': '0.1.5',
    'description': 'my own utils funcs and classes',
    'long_description': None,
    'author': 'Jack Li',
    'author_email': 'lijack1@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import numbers


def gen_cn_price(change_number: numbers.Number):
    """
    .转换数字为大写货币格式( format_word.__len__() - 3 + 2位小数 )
    change_number 支持 float, int, long, string
    """
    format_word = ["分", "角", "元",
                   "拾", "佰", "仟", "万",
                   "拾", "佰", "仟", "亿",
                   "拾", "佰", "仟", "万",
                   "拾", "佰", "仟", "兆"]

    format_num = ["零", "壹", "贰", "叁", "肆", "伍", "陆", "柒", "捌", "玖"]
    if type(change_number) == str:
        # - 如果是字符串,先尝试转换成float或int.
        if '.' in change_number:
            try:
                change_number = float(change_number)
            except:
                '%s can\'t change' % change_number
        else:
            try:
                change_number = int(change_number)
            except:
                '%s   can\'t change' % change_number

    if type(change_number) == float:
        real_numbers = []
        for i in range(len(format_word) - 3, -3, -1):
            if change_number >= 10 ** i or i < 1:
                real_numbers.append(
                    int(round(change_number / (10 ** i), 2) % 10))

    elif isinstance(change_number, int):
        real_numbers = []
        for i in range(len(format_word), -3, -1):
            if change_number >= 10 ** i or i < 1:
                real_numbers.append(
                    int(round(change_number / (10 ** i), 2) % 10))

    else:
        '%s   can\'t change' % change_number

    zflag = 0  # 标记连续0次数，以删除万字，或适时插入零字
    start = len(real_numbers) - 3
    change_words = []
    for i in range(start, -3, -1):  # 使i对应实际位数，负数为角分
        if 0 < real_numbers[start - i] or len(change_words) == 0:
            if zflag:
                change_words.append(format_num[0])
                zflag = 0
            change_words.append(format_num[real_numbers[start - i]])
            change_words.append(format_word[i + 2])

        elif 0 == i or (0 == i % 4 and zflag < 3):  # 控制 万/元
            change_words.append(format_word[i + 2])
            zflag = 0
        else:
            zflag += 1

    if change_words[-1] not in (format_word[0], format_word[1]):
        # - 最后两位非"角,分"则补"整"
        change_words.append("整")

    return ''.join(change_words)
